import os

from grader_service.auth.auth import Authenticator
from grader_service.autograding.kube.kube_grader import KubeAutogradeExecutor
from grader_service.autograding.local_grader import LocalAutogradeExecutor
from grader_service.handlers.base_handler import BaseHandler
from grader_service.orm import User, Lecture
from grader_service.orm.base import DeleteState
from grader_service.orm.lecture import LectureState
from grader_service.orm.takepart import Scope, Role

# from grader_service.autograding.docker_grader import DockerAutogradeExecutor

print("### loading service config")

c.GraderService.service_host = "127.0.0.1"
# existing directory to use as the base directory for the grader service
service_dir = os.path.expanduser("~/grader_service_dir")
c.GraderService.grader_service_dir = service_dir

c.JupyterHubGroupAuthenticator.hub_api_url = "http://127.0.0.1:8081/hub/api"

assert issubclass(KubeAutogradeExecutor, LocalAutogradeExecutor)
# c.RequestHandlerConfig.autograde_executor_class = DockerAutogradeExecutor
c.RequestHandlerConfig.autograde_executor_class = LocalAutogradeExecutor
c.LocalAutogradeExecutor.timeout_func = lambda l: 20

c.KubeAutogradeExecutor.kube_context = "minikube"
c.KubeAutogradeExecutor.resolve_image_name = lambda l, a: "ghcr.io/tu-wien-datalab/grader-labextension:main"

c.CeleryApp.conf = dict(
    broker_url='amqp://localhost',
    result_backend='rpc://',
    task_serializer='json',
    result_serializer='json',
    accept_content=['json'],
    broker_connection_retry_on_startup=True,
    task_always_eager=True
)
c.CeleryApp.worker_kwargs = dict(concurrency=1, pool="prefork")


def get_role_from_auth(auth_state):
    user_role = 'student'
    for role in auth_state['https://purl.imsglobal.org/spec/lti/claim/roles']:
        if role.find('Instructor') >= 1:
            user_role = 'instructor'
            break
    return user_role


def post_auth_hook(authenticator: Authenticator, handler: BaseHandler, authentication: dict):
    print("####### POST AUTH HOOK")
    session = handler.session
    log = handler.log
    auth_state = authentication["auth_state"]

    username = authentication["name"]
    user_model: User = session.query(User).get(username)
    if user_model is None:
        user_model = User()
        user_model.name = username
        session.add(user_model)
        session.commit()

    lecture_code = auth_state["https://purl.imsglobal.org/spec/lti/claim/context"]["label"].replace(" ", "")
    lecture = session.query(Lecture).filter(Lecture.code == lecture_code).one_or_none()
    if lecture is None:
        lecture = Lecture()
        lecture.code = lecture_code
        lecture.name = lecture_code
        lecture.state = LectureState.active
        lecture.deleted = DeleteState.active
        session.add(lecture)
        session.commit()

    lti_role = get_role_from_auth(auth_state)
    scope = Scope[lti_role.lower()]
    log.info(f'Determined role {scope.name} for user {username}')

    role = session.query(Role).filter(Role.username == username, Role.lectid == lecture.id).one_or_none()
    if role is None:
        log.info(f'No role for user {username} in lecture {lecture_code}... creating role')
        role = Role(username=username, lectid=lecture.id, role=scope)
        session.add(role)
        session.commit()
    else:
        log.info(f'Found role {role.role.name} for user {username}  in lecture {lecture_code}... updating role to {scope.name}')
        role.role = scope
        session.commit()

    return authentication


#c.Authenticator.post_auth_hook = post_auth_hook

## authenticator

# JupyterHub client config
c.GraderService.oauth_clients = [{
    'client_id': 'hub',
    'client_secret': 'hub',
    'redirect_uri': 'http://localhost:8080/hub/oauth_callback'
}]

from grader_service.auth.dummy import DummyAuthenticator
from grader_service.auth.pam import PAMAuthenticator
from grader_service.auth.oauth2 import OAuthenticator
from grader_service.auth.lti13.auth import LTI13Authenticator

c.GraderService.authenticator_class = DummyAuthenticator
#
# c.Authenticator.allowed_users = {'instructor', 'tutor', 'student1', 'student2' }
c.Authenticator.allow_all = True

c.DummyAuthenticator.password = 'test'
# c.Authenticator.admin_users = {'user1', 'user2', 'user3', 'user4'}
#
c.GraderService.load_roles = {"lect1:instructor": ["admin", "instructor", "user1"]}

# Moodle/LTI auth

c.LTI13Authenticator.issuer = "http://localhost:8000"
c.LTI13Authenticator.authorize_url = "http://localhost:8000/mod/lti/auth.php"
# The platform's JWKS endpoint url providing public key sets used to verify the ID token
c.LTI13Authenticator.jwks_endpoint = "http://localhost:8000/mod/lti/certs.php"
# The external tool's client id as represented within the platform (LMS)
c.LTI13Authenticator.client_id = ["ItOB9qSSRwH2yDa"]
# calculate username using function defined in username.py
c.LTI13Authenticator.uri_scheme = 'http'



# Copyright (c) 2022, TU Wien
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.
from logging.config import fileConfig
import os
from sqlalchemy import engine_from_config, pool
from alembic import context
import grader_service.orm

# Alembic configuration
config = context.config

# Set up logging from the config file
fileConfig(config.config_file_name)

def configure_db_url():
    """Set the database URL from environment variables if not set in config."""
    if not config.get_main_option('sqlalchemy.url'):
        db_url = os.getenv("GRADER_DB_URL", "sqlite:///grader.db")
        config.set_main_option('sqlalchemy.url', db_url)

configure_db_url()

# Metadata for autogenerate support
target_metadata = grader_service.orm.Base.metadata

def run_migrations_offline():
    """Run migrations in 'offline' mode."""
    url = config.get_main_option("sqlalchemy.url")
    context.configure(
        url=url,
        target_metadata=target_metadata,
        literal_binds=True,
        dialect_opts={"paramstyle": "named"},
        render_as_batch=True
    )

    with context.begin_transaction():
        context.run_migrations()

def run_migrations_online():
    """Run migrations in 'online' mode."""
    connectable = config.attributes.get('connection', None)
    if connectable is None:
        connectable = engine_from_config(
            config.get_section(config.config_ini_section),
            prefix='sqlalchemy.',
            poolclass=pool.NullPool
        )

    try:
        with connectable.connect() as connection:
            context.configure(
                connection=connection,
                target_metadata=target_metadata,
                render_as_batch=True
            )

            with context.begin_transaction():
                context.run_migrations()
    except Exception as e:
        raise

# Run migrations according to the mode
if context.is_offline_mode():
    run_migrations_offline()
else:
    run_migrations_online()


from jupyterhub import orm

## generic
c.JupyterHub.admin_access = True
c.Spawner.default_url = '/lab'
c.Spawner.cmd = ["jupyter-labhub"]

## authenticator
from oauthenticator.generic import GenericOAuthenticator

c.JupyterHub.authenticator_class = GenericOAuthenticator
c.GenericOAuthenticator.oauth_callback_url = "http://localhost:8080/hub/oauth_callback"

c.GenericOAuthenticator.client_id = "hub"
c.GenericOAuthenticator.client_secret = "hub"
c.GenericOAuthenticator.authorize_url = "http://localhost:4010/services/grader/api/oauth2/authorize"
c.GenericOAuthenticator.token_url = "http://localhost:4010/services/grader/api/oauth2/token"

c.GenericOAuthenticator.userdata_url = "http://localhost:4010/services/grader/api/user"
c.GenericOAuthenticator.username_claim = "name"




def post_auth_hook(authenticator, handler, authentication):
    db = authenticator.db
    log = authenticator.log
    username = authentication["name"]
    log.debug(f"{username=:}")
    log.debug(f'auth_state: {authentication["auth_state"]}')
    roles = authentication["auth_state"]["oauth_user"]["roles"]    
    for role in roles:
        lecture_code, lecture_role = role.split(':')
        
        group = orm.Group.find(db, role)
        if group is None:
            log.info(f"Creating group: '{role}'")
            group = orm.Group()
            group.name = role
            db.add(group)
            db.commit()
        
        user_model: orm.User = orm.User.find(db, username)
        if user_model not in group.users:
            log.info(f"Adding user to group: {group.name}")
            group.users.append(user_model)
            db.commit()

        
    return authentication
    

c.Authenticator.post_auth_hook = post_auth_hook
c.Authenticator.enable_auth_state = True

##############################
# PASS GRADER TOKEN TO SPAWNER

def userdata_hook(spawner, auth_state):
    token = auth_state["access_token"]

    # The environment variable GRADER_API_TOKEN is used by the lab-extension
    # to identify the user in API calls to the Grader Service.
    spawner.environment.update({"GRADER_API_TOKEN": token})


# We have access to the authentication data, which we can use to set
# `userdata` in the spawner of the user.
c.Spawner.auth_state_hook = userdata_hook

##############################

# c.Authenticator.allowed_users = {'instructor', 'tutor', 'student1', 'student2'}
c.Authenticator.allow_all = True
c.Authenticator.admin_users = ["admin"]

## spawner
c.JupyterHub.spawner_class = 'jupyterhub.spawner.SimpleLocalProcessSpawner'
c.SimpleLocalProcessSpawner.home_dir_template = '/tmp/lab_dir/{username}'

## simple setup
# c.JupyterHub.ip = '0.0.0.0' # proxy
c.JupyterHub.ip = '127.0.0.1'
c.JupyterHub.port = 8080

## setup with https
# c.JupyterHub.hub_ip = "192.168.5.145"
# c.JupyterHub.ip = '0.0.0.0' # proxy
# c.JupyterHub.port = 8080

# c.JupyterHub.ssl_key  = "jupyterhub_config/dl.key"
# c.JupyterHub.ssl_cert = "jupyterhub_config/__dl_hpc_tuwien_ac_at_cert.cer"


c.JupyterHub.services.append(
    {
        'name': 'grader',
        'url': 'http://127.0.0.1:4010',
        'api_token': '7572f93a2e7640999427d9289c8318c0'
    }
)

c.JupyterHub.log_level = "INFO"

## Is ignored -> actual config is in ~/.jupyter/jupyter_server_config.py
import os

c.GitService.git_access_token = os.environ.get("JUPYTERHUB_API_TOKEN")
c.GitService.git_remote_url = "127.0.0.1:4010/services/grader/git"
c.GitService.git_http_scheme = "http"




import json
import os

import jsonschema
from jsonschema import ValidationError
from nbformat.notebooknode import NotebookNode
from traitlets.config import LoggingConfigurable

root = os.path.dirname(__file__)


class SchemaMismatchError(Exception):
    def __init__(self, message, actual_version, expected_version):
        super(SchemaMismatchError, self).__init__(message)
        self.actual_version = actual_version
        self.expected_version = expected_version


class SchemaTooOldError(SchemaMismatchError):
    pass


class SchemaTooNewError(SchemaMismatchError):
    pass


class BaseMetadataValidator(LoggingConfigurable):
    def __init__(self) -> None:
        with open(
            os.path.join(root, "v{:d}.json".format(self.schema_version)), "r"
        ) as fh:
            self.schema = json.loads(fh.read())

    def _remove_extra_keys(self, cell: NotebookNode) -> None:
        meta = cell.metadata["nbgrader"]
        allowed = set(self.schema["properties"].keys())
        keys = set(meta.keys()) - allowed
        if len(keys) > 0:
            self.log.warning(
                "extra keys detected in metadata, these will be removed: {}".format(
                    keys
                )
            )
            for key in keys:
                del meta[key]

    def upgrade_notebook_metadata(self, nb: NotebookNode) -> NotebookNode:
        for cell in nb.cells:
            self.upgrade_cell_metadata(cell)
        return nb

    def upgrade_cell_metadata(
        self, cell: NotebookNode
    ) -> NotebookNode:  # pragma: no cover
        raise NotImplementedError("this method must be implemented by subclasses")

    def validate_cell(self, cell: NotebookNode) -> None:
        if "nbgrader" not in cell.metadata:
            return
        schema = cell.metadata["nbgrader"].get("schema_version", 0)
        if schema < self.schema_version:
            raise SchemaTooOldError(
                "Outdated schema version: {} (expected {})".format(
                    schema, self.schema_version
                ),
                schema,
                self.schema_version,
            )
        elif schema > self.schema_version:
            raise SchemaTooNewError(
                "Schema version is too new: {} (expected {})".format(
                    schema, self.schema_version
                ),
                schema,
                self.schema_version,
            )
        jsonschema.validate(cell.metadata["nbgrader"], self.schema)

    def validate_nb(self, nb: NotebookNode) -> None:
        for cell in nb.cells:
            self.validate_cell(cell)



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
# import os
# import sys
# sys.path.insert(0, os.path.abspath('../../grader_service'))


# -- Project information -----------------------------------------------------

project = 'Grader Service'
copyright = '2023, Marijana Petojevic, Matthias Matt, Florian Jäger, Elias Wimmer'
author = 'Marijana Petojevic, Matthias Matt, Florian Jäger, Elias Wimmer'

# The full version, including alpha/beta/rc tags
from grader_service._version import __version__
release = __version__

# -- General configuration ---------------------------------------------------

add_module_names = False

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = ["sphinx.ext.autodoc",
              "sphinx.ext.autosummary",
              "sphinx.ext.doctest",
              "sphinx.ext.extlinks",
              "sphinx.ext.viewcode",
              "sphinx.ext.todo",
              "numpydoc",  # handle NumPy documentation formatted docstrings
              "sphinx.ext.intersphinx",
              "sphinx.ext.coverage",
              "sphinx.ext.mathjax",
              "sphinx.ext.ifconfig",]

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

numpydoc_show_class_members = False


# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = []

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = 'pydata_sphinx_theme'

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']

html_css_files = [
    'css/custom.css',
]

html_sidebars = {
    "**": ["search-field.html", "sidebar-nav-bs.html", "sidebar-ethical-ads.html"]
}
